from typing import Iterable, Optional
import polars as pl
import numpy as np
import pandas as pd
from moocore import eaf, eafdiff

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon, Rectangle
import seaborn as sbs

from .metrics import (
    aggegate_running_time,
    get_sequence,
    aggegate_convergence,
    get_tournament_ratings,
    get_attractor_network,
    transform_fval,
)
from .align import align_data, turbo_align
from .indicators import add_indicator, final


matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42
font = {"size": 24}
plt.rc("font", **font)


# tradeoff between simple (few parameters) and flexible. Maybe many parameter but everything with clear defaults?
# Can also make sure any useful function for data processing is available separately for more flexibility


def single_function_fixedtarget(
    data: pl.DataFrame,
    evaluation_variable: str = "evaluations",
    fval_variable: str = "raw_y",
    free_variables: Iterable[str] = ["algorithm_name"],
    f_min: float = None,
    f_max: float = None,
    max_budget: int = None,
    maximization: bool = False,
    measures: Iterable[str] = ["ERT"],
    scale_xlog: bool = True,
    scale_ylog: bool = True,
    ax: matplotlib.axes._axes.Axes = None,
    file_name: str = None,
):
    """Create a fixed-target plot for a given set of performance data.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should        be generated from a DataManager.
        evaluation_variable (str, optional): Column in 'data' which corresponds to the number of evaluations. Defaults to "evaluations".
        fval_variable (str, optional): Column in 'data' which corresponds to the performance measure. Defaults to "raw_y".
        free_variables (Iterable[str], optional): Columns in 'data' which correspond to the variables which will be used to distinguish between lines in the plot. Defaults to ["algorithm_name"].
        f_min (float, optional): Minimum value to use for the 'fval_variable', if not present the min of that column will be used. Defaults to None.
        f_max (float, optional): Maximum value to use for the 'fval_variable', if not present the max of that column will be used. Defaults to None.
        max_budget (int, optional): Maximum value to use for the 'evaluation_variable', if not present the max of that column will be used. Defaults to None.
        maximization (bool, optional): Boolean indicating whether the 'fval_variable' is being maximized. Defaults to False.
        measures (Iterable[str], optional): List of measures which should be used in the plot. Valid options are 'ERT', 'mean', 'PAR-10', 'min', 'max'. Defaults to ['ERT'].
        scale_xlog (bool, optional): Should the x-axis be log-scaled. Defaults to True.
        scale_ylog (bool, optional): Should the y-axis be log-scaled. Defaults to True.
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.

    Returns:
        pd.DataFrame: The final dataframe which was used to create the plot
    """
    dt_agg = aggegate_running_time(
        data,
        evaluation_variable=evaluation_variable,
        fval_variable=fval_variable,
        free_variables=free_variables,
        f_min=f_min,
        f_max=f_max,
        scale_flog=scale_xlog,
        max_budget=max_budget,
        maximization=maximization,
    )

    dt_molt = dt_agg.melt(id_vars=[fval_variable] + free_variables)
    dt_plot = dt_molt[dt_molt["variable"].isin(measures)].sort_values(free_variables)

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(16, 9))
    sbs.lineplot(
        dt_plot,
        x=fval_variable,
        y="value",
        style="variable",
        hue=dt_plot[free_variables].apply(tuple, axis=1),
        ax=ax,
    )
    if scale_xlog:
        ax.set_xscale("log")
    if scale_ylog:
        ax.set_yscale("log")

    if not maximization:
        ax.set_xlim(ax.get_xlim()[::-1])

    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)

    return dt_plot


def single_function_fixedbudget(
    data: pl.DataFrame,
    evaluation_variable: str = "evaluations",
    fval_variable: str = "raw_y",
    free_variables: Iterable[str] = ["algorithm_name"],
    x_min: float = None,
    x_max: float = None,
    maximization: bool = False,
    measures: Iterable[str] = ["geometric_mean"],
    scale_xlog: bool = True,
    scale_ylog: bool = True,
    ax: matplotlib.axes._axes.Axes = None,
    file_name: str = None,
):
    """Create a fixed-budget plot for a given set of performance data.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should        be generated from a DataManager.
        evaluation_variable (str, optional): Column in 'data' which corresponds to the number of evaluations. Defaults to "evaluations".
        fval_variable (str, optional): Column in 'data' which corresponds to the performance measure. Defaults to "raw_y".
        free_variables (Iterable[str], optional): Columns in 'data' which correspond to the variables which will be used to distinguish between lines in the plot. Defaults to ["algorithm_name"].
        x_min (float, optional): Minimum value to use for the 'evaluation_variable', if not present the min of that column will be used. Defaults to None.
        x_max (float, optional): Maximum value to use for the 'evaluation_variable', if not present the max of that column will be used. Defaults to None.
        maximization (bool, optional): Boolean indicating whether the 'fval_variable' is being maximized. Defaults to False.
        measures (Iterable[str], optional): List of measures which should be used in the plot. Valid options are 'geometric_mean', 'mean', 'median', 'min', 'max'. Defaults to ['geometric_mean'].
        scale_xlog (bool, optional): Should the x-axis be log-scaled. Defaults to True.
        scale_ylog (bool, optional): Should the y-axis be log-scaled. Defaults to True.
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.

    Returns:
        pd.DataFrame: The final dataframe which was used to create the plot
    """
    dt_agg = aggegate_convergence(
        data,
        evaluation_variable=evaluation_variable,
        fval_variable=fval_variable,
        free_variables=free_variables,
        x_min=x_min,
        x_max=x_max,
        maximization=maximization,
    )

    dt_molt = dt_agg.melt(id_vars=[evaluation_variable] + free_variables)
    dt_plot = dt_molt[dt_molt["variable"].isin(measures)].sort_values(free_variables)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(16, 9))
    sbs.lineplot(
        dt_plot,
        x=evaluation_variable,
        y="value",
        style="variable",
        hue=dt_plot[free_variables].apply(tuple, axis=1),
        ax=ax,
    )
    if scale_xlog:
        ax.set_xscale("log")
    if scale_ylog:
        ax.set_yscale("log")


    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)

    return dt_plot


def heatmap_single_run(
    data: pl.DataFrame,
    var_cols: Iterable[str],
    eval_col: str = "evaluations",
    scale_xlog: bool = True,
    x_mins: Iterable[float] = [-5],
    x_maxs: Iterable[float] = [5],
    ax: matplotlib.axes._axes.Axes = None,
    file_name: Optional[str] = None,
):
    """Create a heatmap showing the search space points evaluated in a single run

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        var_cols (Iterable[str]): The variables which correspond to the searchspace variable columns
        eval_col (str): The variable corresponding to evaluations. Defaults to 'evaluations'
        scale_xlog (bool, optional): Whether the evaluations should be log-scaled. Defaults to True.
        x_mins (Iterable[float], optional): Minimum bound for the variables. Should be of the same length as 'var_cols'. Defaults to [-5].
        x_maxs (Iterable[float], optional): Maximum bound for the variables. Should be of the same length as 'var_cols'.. Defaults to [5].
        ax (matplotlib.axes._axes.Axes, optional): Axis on which to create the plot. Defaults to None.
        file_name (Optional[str], optional): If ax is not given, filename to save the plot. Defaults to None.

    Returns:
        pd.DataFrame: pandas dataframe of the exact data used to create the plot
    """
    assert data["data_id"].n_unique() == 1
    dt_plot = data[var_cols].transpose().to_pandas()
    dt_plot.columns = list(data["evaluations"])
    dt_plot = (dt_plot - x_mins) / (x_maxs - x_mins)
    if ax is None:
        fig, ax = plt.subplots(figsize=(32, 9))
    sbs.heatmap(dt_plot, cmap="viridis", vmin=0, vmax=1, ax=ax)
    if scale_xlog:
        ax.set_xscale("log")
        ax.set_xlim(1, len(data))

    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)
    return dt_plot


def plot_eaf_singleobj(
    data: pl.DataFrame,
    min_budget: int = None,
    max_budget: int = None,
    scale_xlog: bool = True,
    n_quantiles: int = 100,
    eval_var: str = "evaluations",
    fval_var: str = "raw_y",
    ax: matplotlib.axes._axes.Axes = None,
    file_name: Optional[str] = None,
):
    """Plot the EAF for a single objective column agains budget. For the EAF-plot for multiple objective
    columns, see 'plot_eaf_pareto'.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        n_quantiles (int, optional): Number of discrete levels in the EAF. Defaults to 100.
        eval_var (str, optional): The variable corresponding to evaluations. Defaults to 'evaluations'
        fval_var (str, optional): The variable corresponding to function values. Defaults to "raw_y".
        scale_xlog (bool, optional): Whether the evaluations should be log-scaled. Defaults to True.
        min_budget (Iterable[float], optional): Minimum bound for the variables. Should be of the same length as 'var_cols'. Defaults to [-5].
        max_budget (Iterable[float], optional): Maximum bound for the variables. Should be of the same length as 'var_cols'.. Defaults to [5].
        ax (matplotlib.axes._axes.Axes, optional): Axis on which to create the plot. Defaults to None.
        file_name (Optional[str], optional): If ax is not given, filename to save the plot. Defaults to None.

    Returns:
        pd.DataFrame: pandas dataframe of the exact data used to create the plot
    """
    if min_budget is None:
        min_budget = data[eval_var].min()
    if max_budget is None:
        max_budget = data[eval_var].max()
    evals = get_sequence(min_budget, max_budget, 50, scale_xlog, True)
    long = align_data(data, np.array(evals, "uint64"), ["data_id"], output="long")

    quantiles = np.arange(0, 1 + 1 / ((n_quantiles - 1) * 2), 1 / (n_quantiles - 1))
    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 9))
    colors = sbs.color_palette("viridis", n_colors=len(quantiles))
    for quant, color in zip(quantiles, colors[::-1]):
        poly = np.array(
            long.group_by(eval_var).quantile(quant).sort(eval_var)[eval_var, fval_var]
        )
        poly = np.append(
            poly, np.array([[max(poly[:, 0]), long[fval_var].max()]]), axis=0
        )
        poly = np.append(
            poly, np.array([[min(poly[:, 0]), long[fval_var].max()]]), axis=0
        )
        poly2 = np.repeat(poly, 2, axis=0)
        poly2[2::2, 1] = poly[:, 1][:-1]
        ax.add_patch(Polygon(poly2, facecolor=color))
    ax.set_ylim(long[fval_var].min(), long[fval_var].max())
    ax.set_xlim(min(evals), max(evals))
    ax.set_axisbelow(True)
    ax.grid(which="both", zorder=100)
    ax.set_yscale("log")
    if scale_xlog:
        ax.set_xscale("log")

    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)
    return long


def plot_eaf_pareto(
    data: pl.DataFrame,
    x_column: str,
    y_column: str,
    min_y: float = 0,
    max_y: float = 1,
    scale_xlog: bool = False,
    scale_ylog: bool = False,
    ax: matplotlib.axes._axes.Axes = None,
    filename_fig: Optional[str] = None,
):
    """Plot the EAF for two arbitrary data columns. For the EAF-plot for single-objective
    optimization runs, the 'plot_eaf_singleobj' provides a simpler interface.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        x_column (str, optional): The variable corresponding to the first objective.
        y_column (str, optional): The variable corresponding to the second objective.
        min_y (float): Minimum value for the second objective.
        max_y (float): Maximum value for the second objective.
        scale_xlog (bool, optional): Whether the first objective should be log-scaled. Defaults to False.
        scale_ylog (bool, optional): Whether the second objective should be log-scaled. Defaults to False.
        ax (matplotlib.axes._axes.Axes, optional): Axis on which to create the plot. Defaults to None.
        file_name (Optional[str], optional): If ax is not given, filename to save the plot. Defaults to None.
    """

    eaf_data = eaf(np.array(data[[x_column, y_column, "data_id"]]))
    eaf_data_df = pd.DataFrame(eaf_data)
    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 9))
    colors = sbs.color_palette("viridis", n_colors=eaf_data_df[2].nunique())
    eaf_data_df = eaf_data_df.sort_values(0)
    min_x = np.min(eaf_data_df[0])
    max_x = np.max(eaf_data_df[0])
    if min_y is None:
        min_y = np.min(eaf_data_df[1])
    if max_y is None:
        max_y = np.max(eaf_data_df[1])
    for i, color in zip(eaf_data_df[2].unique(), colors[::-1]):
        poly = np.array(eaf_data_df[eaf_data_df[2] == i][[0, 1]])
        # poly = np.append(poly, np.array([[max(poly[:, 0]), max(poly[:, 1])]]), axis=0)
        # poly = np.append(poly, np.array([[min(poly[:, 0]), max(poly[:, 1])]]), axis=0)
        poly = np.append(poly, np.array([[max_x, max_y]]), axis=0)
        poly = np.append(poly, np.array([[min(poly[:, 0]), max_y]]), axis=0)
        poly2 = np.repeat(poly, 2, axis=0)
        poly2[2::2, 1] = poly[:, 1][:-1]
        ax.add_patch(Polygon(poly2, facecolor=color))
    # ax.add_colorbar()
    ax.set_ylim(min_y, max_y)
    ax.set_xlim(min_x, max_x)
    ax.set_axisbelow(True)
    sm = plt.cm.ScalarMappable(cmap="viridis", norm=plt.Normalize(vmin=0, vmax=1))
    sm.set_array([])
    plt.colorbar(sm, ax=ax)
    if scale_ylog:
        ax.set_yscale("log")
    if scale_xlog:
        ax.set_xscale("log")
    ax.grid(which="both", zorder=100)
    if filename_fig:
        fig.tight_layout()
        fig.savefig(filename_fig)


def eaf_diffs(
    data1: pl.DataFrame,
    data2: pl.DataFrame,
    x_column: str,
    y_column: str,
    min_y: float = 0,
    max_y: float = 1,
    scale_xlog: bool = False,
    scale_ylog: bool = False,
    ax: matplotlib.axes._axes.Axes = None,
    filename_fig: Optional[str] = None,
):
    """Plot the EAF differences between two datasets.

    Args:
        data1 (pl.DataFrame): The DataFrame which contains the full performance trajectory for algorithm 1. Should be generated from a DataManager.
        data2 (pl.DataFrame): The DataFrame which contains the full performance trajectory for algorithm 2. Should be generated from a DataManager.
        x_column (str, optional): The variable corresponding to the first objective.
        y_column (str, optional): The variable corresponding to the second objective.
        min_y (float): Minimum value for the second objective.
        max_y (float): Maximum value for the second objective.
        scale_xlog (bool, optional): Whether the first objective should be log-scaled. Defaults to False.
        scale_ylog (bool, optional): Whether the second objective should be log-scaled. Defaults to False.
        ax (matplotlib.axes._axes.Axes, optional): Axis on which to create the plot. Defaults to None.
        file_name (Optional[str], optional): If ax is not given, filename to save the plot. Defaults to None.
    """
    # TODO: add an approximation version to speed up plotting
    x = np.array(data1[[x_column, y_column, "data_id"]])
    y = np.array(data2[[x_column, y_column, "data_id"]])
    eaf_diff_rect = eafdiff(x, y, rectangles=True)
    color_dict = {
        k: v
        for k, v in zip(
            np.unique(eaf_diff_rect[:, -1]),
            sbs.color_palette("viridis", n_colors=len(np.unique(eaf_diff_rect[:, -1]))),
        )
    }
    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 9))
    for rect in eaf_diff_rect:
        ax.add_patch(
            Rectangle(
                (rect[0], rect[1]),
                rect[2] - rect[0],
                rect[3] - rect[1],
                facecolor=color_dict[rect[-1]],
            )
        )
    if min_y is None:
        min_y = np.min(x[1])
    if max_y is None:
        max_y = np.max(x[1])
    ax.set_ylim(min_y, max_y)
    if scale_ylog:
        ax.set_yscale("log")
    if scale_xlog:
        ax.set_xscale("log")
    if filename_fig:
        fig.tight_layout()
        fig.savefig(filename_fig)


def plot_ecdf(
    data,
    fval_var: str = "raw_y",
    eval_var: str = "evaluations",
    free_vars: Iterable[str] = ["algorithm_name"],
    scale_xlog: bool = True,
    x_min: int = None,
    x_max: int = None,
    x_values: Iterable[int] = None,
    y_min: int = None,
    y_max: int = None,
    scale_ylog: bool = True,
    maximization: bool = False,
    ax: matplotlib.axes._axes.Axes = None,
    file_name: Optional[str] = None,
):
    """Function to plot empirical cumulative distribution function (Based on EAF)

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        eval_var (str, optional): Column in 'data' which corresponds to the number of evaluations. Defaults to "evaluations".
        fval_var (str, optional): Column in 'data' which corresponds to the performance measure. Defaults to "raw_y".
        free_vars (Iterable[str], optional): Columns in 'data' which correspond to the variables which will be used to distinguish between lines in the plot. Defaults to ["algorithm_name"].
        x_min (float, optional): Minimum value to use for the 'eval_var', if not present the min of that column will be used. Defaults to None.
        x_max (float, optional): Maximum value to use for the 'eval_var', if not present the max of that column will be used. Defaults to None.
        x_values (Iterable[int], optional): List of x-values at which to plot the ECDF. If not provided, the x_min, x_max and scale_xlog arguments will be used to sample these points.
        scale_xlog (bool, optional): Should the x-axis be log-scaled. Defaults to True.
        y_min (float, optional): Minimum value to use for the 'fval_var', if not present the min of that column will be used. Defaults to None.
        y_max (float, optional): Maximum value to use for the 'fval_var', if not present the max of that column will be used. Defaults to None.
        scale_ylog (bool, optional): Should the y-values be log-scaled before normalization. Defaults to True.
        maximization (bool, optional): Boolean indicating whether the 'fval_var' is being maximized. Defaults to False.
        measures (Iterable[str], optional): List of measures which should be used in the plot. Valid options are 'geometric_mean', 'mean', 'median', 'min', 'max'. Defaults to ['geometric_mean'].
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.

    Returns:
        pd.DataFrame: pandas dataframe of the exact data used to create the plot
    """
    if x_min is None:
        x_min = data[eval_var].min()
    if x_max is None:
        x_max = data[eval_var].max()
    x_values = get_sequence(x_min, x_max, 50, scale_log=scale_xlog, cast_to_int=True)

    data_aligned = turbo_align(
        data,
        x_values,
        x_col=eval_var,
        y_col=fval_var,
        maximization=maximization,
    )
    dt_plot = (
        transform_fval(data_aligned, fval_col=fval_var, maximization=maximization)
        .group_by([eval_var] + free_vars)
        .mean()
        .sort(eval_var)
    ).to_pandas()

    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 9))
    if len(free_vars) == 1:
        hue_arg = free_vars[0]
        style_arg = free_vars[0]
    else:
        style_arg = free_vars[0]
        hue_arg = dt_plot[free_vars[1:]].apply(tuple, axis=1)

    sbs.lineplot(
        dt_plot,
        x="evaluations",
        y="eaf",
        style=style_arg,
        hue=hue_arg,
        ax=ax,
    )
    if scale_xlog:
        ax.set_xscale("log")
    ax.grid()
    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)
    return dt_plot


def multi_function_fixedbudget():
    # either just loop over function column(s), or more advanced
    raise NotImplementedError


def multi_function_fixedtarget():
    raise NotImplementedError


def plot_tournament_ranking(
    data,
    alg_vars: Iterable[str] = ["algorithm_name"],
    fid_vars: Iterable[str] = ["function_name"],
    perf_var: str = "raw_y",
    nrounds: int = 25,
    maximization: bool = False,
    ax: matplotlib.axes._axes.Axes = None,
    file_name: str = None,
):
    """Method to plot ELO ratings of a set of algorithm on a set of problems.
    Calculated based on nrounds of competition, where in each round all algorithms face all others (pairwise) on every function.
    For each round, a sampled performance value is taken from the data and used to determine the winner.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        alg_vars (Iterable[str], optional): Which variables specific the algortihms which will compete. Defaults to ["algorithm_name"].
        fid_vars (Iterable[str], optional): Which variables denote the problems on which will be competed. Defaults to ["function_name"].
        perf_var (str, optional): Which variable corresponds to the performance. Defaults to "raw_y".
        nrounds (int, optional): How many round should be played. Defaults to 25.
        maximization (bool, optional): Whether the performance should be maximized. Defaults to False.
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.

    Returns:
        pd.DataFrame: pandas dataframe of the exact data used to create the plot
    """
    # candlestick plot based on average and volatility
    dt_elo = get_tournament_ratings(
        data, alg_vars, fid_vars, perf_var, nrounds, maximization
    )
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(10, 5))

    sbs.pointplot(data=dt_elo, x=alg_vars[0], y="Rating", linestyle="none", ax=ax)

    ax.errorbar(
        dt_elo[alg_vars[0]],
        dt_elo["Rating"],
        yerr=dt_elo["Deviation"],
        fmt="o",
        color="blue",
        alpha=0.6,
        capsize=5,
        elinewidth=1.5,
    )
    ax.grid()

    if file_name:
        plt.tight_layout()
        plt.savefig(file_name)
    return dt_elo


def robustranking():
    # to decide which plot(s) to use and what exact interface to define
    raise NotImplementedError()


def stats_comparison():
    # heatmap or graph of statistical comparisons
    raise NotImplementedError()


def winnning_fraction_heatmap():
    # nevergrad-like heatmap
    raise NotImplementedError()


def plot_paretofronts_2d(
    data: pl.DataFrame,
    obj_vars: Iterable[str] = ["raw_y", "F2"],
    free_vars: Iterable[str] = ["algorithm_name"],
    ax: matplotlib.axes._axes.Axes = None,
    file_name: str = None,
):
    """Very basic plot to visualize pareto fronts

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        obj_vars (Iterable[str], optional): Which variables (length should be 2) to use for plotting. Defaults to ["raw_y", "F2"].
        free_vars (Iterable[str], optional): Which varialbes should be used to distinguish between categories. Defaults to ["algorithm_name"].
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.

    Returns:
        pd.DataFrame: pandas dataframe of the exact data used to create the plot
    """
    assert len(obj_vars) == 2

    df = add_indicator(data, final.NonDominated(), obj_vars)

    if ax is None:
        fig, ax = plt.subplots(figsize=(16, 9))
    sbs.scatterplot(df, x=data.obj_vars[0], y=data.obj_vars[1], hue=free_vars, ax=ax)
    if ax is None and file_name:
        fig.tight_layout()
        fig.savefig(file_name)
    return df


def plot_indicator_over_time(
    data: pl.DataFrame,
    obj_columns: Iterable[str],
    indicator: object,
    eval_column: str = "evaluations",
    evals_min: int = 0,
    evals_max: int = 50_000,
    nr_eval_steps: int = 50,
    free_variable: str = "algorithm_name",
    ax: matplotlib.axes._axes.Axes = None,
    filename_fig: Optional[str] = None,
):
    """Convenience function to plot the anytime performance of a single indicator.

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        obj_columns (Iterable[str], optional): Which columns in 'data' correspond to the objectives.
        indicator (object): Indicator object from iohinspector.indicators
        eval_column (Iterable[str], optional): Which columns in 'data' correspond to the objectives. Defaults to 'evaluations'.
        evals_min (int, optional): Lower bound for eval_column. Defaults to 0.
        evals_max (int, optional): Upper bound for eval_column. Defaults to 50_000.
        nr_eval_steps (int, optional): Number of steps between lower and upper bounds of eval_column. Defaults to 50.
        free_variable (str, optional): Variable which corresponds to category to differentiate in the plot. Defaults to 'algorithm_name'.
        ax (matplotlib.axes._axes.Axes, optional): Existing matplotlib axis object to draw the plot on.
        file_name (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.
    """

    evals = get_sequence(
        evals_min, evals_max, nr_eval_steps, cast_to_int=True, scale_log=True
    )
    df = add_indicator(
        data, indicator, objective_columns=obj_columns, evals=evals
    ).to_pandas()
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(16, 9))
    sbs.lineplot(
        df,
        x=eval_column,
        y=indicator.var_name,
        hue=free_variable,
        palette=sbs.color_palette(n_colors=len(np.unique(data[free_variable]))),
        ax=ax,
    )
    ax.set_xlabel(eval_column)
    ax.set_xlim(evals_min, evals_max)
    ax.set_xscale("log")
    ax.grid()
    if filename_fig:
        fig.tight_layout()
        fig.savefig(filename_fig)

    return df


def plot_robustrank_over_time(
    data: pl.DataFrame,
    obj_columns: Iterable[str],
    evals: Iterable[int],
    indicator: object,
    filename_fig: Optional[str] = None,
):
    """Plot robust ranking at distinct timesteps

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        obj_columns (Iterable[str], optional): Which columns in 'data' correspond to the objectives.
        evals (Iterable[int]): Timesteps at which to get the rankings
        indicator (object): Indicator object from iohinspector.indicators
        filename_fig (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.
    """
    from robustranking import Benchmark
    from robustranking.comparison import MOBootstrapComparison, BootstrapComparison
    from robustranking.utils.plots import plot_ci_list, plot_line_ranks

    df = add_indicator(
        data, indicator, objective_columns=obj_columns, evals=evals
    ).to_pandas()
    df_part = df[["evaluations", indicator.var_name, "algorithm_name", "run_id"]]
    dt_pivoted = pd.pivot(
        df_part,
        index=["algorithm_name", "run_id"],
        columns=["evaluations"],
        values=[indicator.var_name],
    ).reset_index()
    dt_pivoted.columns = ["algorithm_name", "run_id"] + evals
    benchmark = Benchmark()
    benchmark.from_pandas(dt_pivoted, "algorithm_name", "run_id", evals)

    comparison = MOBootstrapComparison(
        benchmark,
        alpha=0.05,
        minimise=indicator.minimize,
        bootstrap_runs=1000,
        aggregation_method=np.mean,
    )
    fig, axs = plt.subplots(1, 4, figsize=(16, 9), sharey=True)
    for ax, runtime in zip(axs.ravel(), benchmark.objectives):
        plot_ci_list(comparison, objective=runtime, ax=ax)
        if runtime != evals[0]:
            ax.set_ylabel("")
        if runtime != evals[-1]:
            ax.get_legend().remove()
        ax.set_title(runtime)

    plt.tight_layout()
    if filename_fig:
        plt.savefig(filename_fig)
        plt.close()


def plot_robustrank_changes(
    data: pl.DataFrame,
    obj_columns: Iterable[str],
    evals: Iterable[int],
    indicator: object,
    filename_fig: Optional[str] = None,
):
    """Plot robust ranking changes at distinct timesteps

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        obj_columns (Iterable[str], optional): Which columns in 'data' correspond to the objectives.
        evals (Iterable[int]): Timesteps at which to get the rankings
        indicator (object): Indicator object from iohinspector.indicators
        filename_fig (str, optional): Where should the resulting plot be stored. Defaults to None. If existing axis is provided, this functionality is disabled.
    """
    from robustranking import Benchmark
    from robustranking.comparison import MOBootstrapComparison, BootstrapComparison
    from robustranking.utils.plots import plot_ci_list, plot_line_ranks

    df = add_indicator(
        data, indicator, objective_columns=obj_columns, evals=evals
    ).to_pandas()
    df_part = df[["evaluations", indicator.var_name, "algorithm_name", "run_id"]]
    dt_pivoted = pd.pivot(
        df_part,
        index=["algorithm_name", "run_id"],
        columns=["evaluations"],
        values=[indicator.var_name],
    ).reset_index()
    dt_pivoted.columns = ["algorithm_name", "run_id"] + evals

    comparisons = {
        f"{eval}": BootstrapComparison(
            Benchmark().from_pandas(dt_pivoted, "algorithm_name", "run_id", eval),
            alpha=0.05,
            minimise=indicator.minimize,
            bootstrap_runs=1000,
        )
        for eval in evals
    }

    fig, ax = plt.subplots(1, 1, figsize=(16, 9))
    plot_line_ranks(comparisons, ax=ax)

    plt.tight_layout()
    if filename_fig:
        plt.savefig(filename_fig)
        plt.close()


def plot_attractor_network(
    data,
    coord_vars: Iterable[str] = ["x1", "x2"],
    fval_var: str = "raw_y",
    eval_var: str = "evaluations",
    maximization: bool = False,
    beta=40,
    epsilon=0.0001,
):
    """Plot an attractor network from the provided data

    Args:
        data (pl.DataFrame): The original dataframe, should contain the performance and position information
        coord_vars (Iterable[str], optional): Which columns correspond to position information. Defaults to ['x1', 'x2'].
        fval_var (str, optional): Which column corresponds to performance. Defaults to 'raw_y'.
        eval_var (str, optional): Which column corresponds to evaluations. Defaults to 'evaluations'.
        maximization (bool, optional): Whether fval_var is to be maximized. Defaults to False.
        beta (int, optional): Minimum stagnation lenght. Defaults to 40.
        epsilon (float, optional): Radius below which positions should be considered identical in the network. Defaults to 0.0001.

    Returns:
        pd.DataFrame, pd.DataFrame: two dataframes containing the nodes and edges of the network respectively.
    """
    try:
        import networkx as nx
    except:
        print("NetworkX is required to use this plot type")
        return
    from sklearn.decomposition import MDS

    nodes, edges = get_attractor_network(
        data, maximization, coord_vars, fval_var, eval_var, beta, epsilon
    )
    network = nx.DiGraph()
    for idx, row in nodes.iterrows():
        network.add_node(
            idx,
            decision=np.array(row)[: len(coord_vars)],
            fitness=row["y"],
            hitcount=row["count"],
            evals=row["evals"] / row["count"],
        )

    for _, row in edges.iterrows():
        network.add_edge(
            row["start"],
            row["end"],
            weight=row["count"],
            evaldiff=row["stag_length_avg"],
        )
    network.remove_edges_from(nx.selfloop_edges(network))

    decision_matrix = [network.nodes[node]["decision"] for node in network.nodes()]
    mds = MDS(n_components=1, random_state=0)
    x_positions = mds.fit_transform(
        decision_matrix
    ).flatten()  # Flatten to get 1D array for x-axis
    y_positions = [network.nodes[node]["fitness"] for node in network.nodes()]
    pos = {
        node: (x, y) for node, x, y in zip(network.nodes(), x_positions, y_positions)
    }

    hitcounts = [network.nodes[node]["hitcount"] for node in network.nodes()]
    if len(hitcounts) > 1:
        min_hitcount = min(hitcounts)
        max_hitcount = max(hitcounts)
    # Node sizes and colors based on fitness values (as in your original code)
    if len(hitcounts) > 1 and np.std(hitcounts) > 0:
        node_sizes = [
            100
            + (
                400
                * (network.nodes[node]["hitcount"] - min_hitcount)
                / (max_hitcount - min_hitcount)
            )
            for node in network.nodes()
        ]
    else:
        node_sizes = [500] * len(hitcounts)
    fitness_values = y_positions  # Reuse y_positions as they represent 'fitness'
    norm = plt.Normalize(min(fitness_values), max(fitness_values))
    node_colors = plt.cm.viridis(norm(fitness_values))

    # Draw the graph
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 6))
    nx.draw(
        network,
        pos=pos,
        with_labels=False,
        node_size=node_sizes,
        node_color=node_colors[:, :3],
        edge_color="gray",
        width=2,
        ax=ax,
    )

    # Add colorbar for fitness values
    sm = plt.cm.ScalarMappable(cmap="viridis", norm=norm)
    sm.set_array(fitness_values)
    plt.xlabel("MDS-reduced decision vector")
    plt.ylabel("fitness")
    plt.tight_layout()



from typing import Iterable, Callable
from functools import partial

import polars as pl
import numpy as np

from .anytime import *
from .final import *

def add_indicator(
    df: pl.DataFrame, indicator: Callable, objective_columns: Iterable, **kwargs
) -> pl.DataFrame:
    """Adds an indicator to a Polars DataFrame.

    This function applies a specified indicator function to groups of data
    within a Polars DataFrame, grouped by the "data_id" column. The indicator
    is applied to the specified objective columns, along with any additional
    keyword arguments.

    Args:
        df (pl.DataFrame):
            The Polars DataFrame containing the data to which the indicator
            will be added. It must contain a column named "data_id" for
            grouping purposes.
        indicator (object):
            A callable object (e.g., a function) that computes the desired
            indicator. It must accept `objective_columns` and any additional
            arguments passed via `kwargs`.
        objective_columns (Iterable):
            An iterable of column names in the DataFrame that the indicator
            function will process.
        **kwargs:
            Additional keyword arguments to be passed to the `indicator`
            callable.

    Returns:
        pl.DataFrame:
            A new Polars DataFrame with the computed indicator applied to each
            group of data.
    """
    indicator_callable = partial(
        indicator, objective_columns=objective_columns, **kwargs
    )
    return df.group_by("data_id").map_groups(indicator_callable)



import os
import warnings
from dataclasses import dataclass, field
from copy import deepcopy

import polars as pl

from .data import Dataset, Function, Algorithm, METADATA_SCHEMA
from glob import glob

@dataclass
class DataManager:
    data_sets: list[Dataset] = field(default_factory=list, repr=None)
    overview: pl.DataFrame = field(
        default_factory=lambda: pl.DataFrame(schema=METADATA_SCHEMA)
    )

    def __post_init__(self):
        for data_set in self.data_sets:
            self.extend_overview(data_set)
            
    def add_folders(self, folders: list[str]):
        """Utility loop for adding multiple folders"""
        
        for folder in folders:
            self.add_folder(folder)

    def add_folder(self, folder_name: str):
        """Add a folder with ioh generated data"""

        if not os.path.isdir(folder_name):
            raise FileNotFoundError(f"{folder_name} not found")

        json_files = glob(f"{folder_name}/**/*.json", recursive = True)
        if not any(json_files):
            raise FileNotFoundError(f"{folder_name} does not contain any json files")

        datasets = [Dataset.from_json(json_file) for json_file in json_files]

        for ds in datasets:
            self.data_sets.append(ds)

        ds_overviews = pl.concat([ds.overview for ds in datasets], how='diagonal_relaxed')
        self.overview = pl.concat([ds_overviews, self.overview], how='diagonal_relaxed')

    def add_json(self, json_file: str):
        """Add a single json file with ioh generated data"""

        if any((d.file == json_file) for d in self.data_sets):
            warnings.warn(
                f"{json_file} is already loaded. Skipping file", RuntimeWarning
            )
            return
        data_set = Dataset.from_json(json_file)
        self.add_data_set(data_set)

    def extend_overview(self, data_set: Dataset):
        """ "Include a new data set in the manager"""
        columns = set(self.overview.schema.keys())
        ds_columns = set(data_set.overview.schema.keys())

        new_columns = list(ds_columns - columns)
        missing_columns = list(columns - ds_columns)

        ds_overview = data_set.overview.with_columns(
            *(
                pl.lit(None, dtype=self.overview.schema[name]).alias(name)
                for name in missing_columns
            )
        )
        self.overview = self.overview.with_columns(
            *(
                pl.lit(None, dtype=data_set.overview.schema[name]).alias(name)
                for name in new_columns
            )
        )
        self.overview = self.overview.extend(ds_overview.select(self.overview.columns))

    def add_data_set(self, data_set: Dataset):
        """Only use this to add data sets"""
        self.data_sets.append(data_set)
        self.extend_overview(data_set)

    def __add__(self, other: "DataManager") -> "DataManager":
        # TODO: filter on overlap
        return DataManager(deepcopy(self.data_sets) + deepcopy(other.data_sets))

    def select(
        self,
        data_ids: list[int] = None,
        function_ids: list[int] = None,
        algorithms: list[str] = None,
        experiment_attributes: list[tuple[str, str]] = None,
        data_attributes: list[str] = None,
        dimensions: list[int] = None,
        instances: list[int] = None,
    ) -> "DataManager":

        selected_data_sets = deepcopy(self.data_sets)

        if data_ids is not None:
            for dset in selected_data_sets:
                for scen in dset.scenarios:
                    scen.runs = [run for run in scen.runs if run.data_id in data_ids]

                dset.scenarios = [scen for scen in dset.scenarios if any(scen.runs)]
            return DataManager([x for x in selected_data_sets if any(x.scenarios)])

        ## dataset filters
        if function_ids is not None:
            selected_data_sets = [
                x for x in selected_data_sets if x.function.id in function_ids
            ]

        if algorithms is not None:
            selected_data_sets = [
                x for x in selected_data_sets if x.algorithm.name in algorithms
            ]

        if experiment_attributes is not None:
            for attr in experiment_attributes:
                selected_data_sets = [
                    x for x in selected_data_sets if attr in x.experiment_attributes
                ]

        if data_attributes is not None:
            for attr in data_attributes:
                selected_data_sets = [
                    x for x in selected_data_sets if attr in x.data_attributes
                ]

        ## scenario_filters
        if dimensions is not None:
            for dset in selected_data_sets:
                dset.scenarios = [
                    scen for scen in dset.scenarios if scen.dimension in dimensions
                ]

        ## run filter
        if instances is not None:
            for dset in selected_data_sets:
                for scen in dset.scenarios:
                    scen.runs = [run for run in scen.runs if run.instance in instances]

        return DataManager(selected_data_sets)

    def select_indexes(self, idxs):
        return DataManager([self.data_sets[idx] for idx in idxs])

    @property
    def functions(self) -> tuple[Function]:
        return tuple([x.function for x in self.data_sets])

    @property
    def algorithms(self) -> tuple[Algorithm]:
        return tuple([x.algorithm for x in self.data_sets])

    @property
    def experiment_attributes(self) -> tuple[tuple[str, str]]:
        attrs = []
        for data_set in self.data_sets:
            for attr in data_set.experiment_attributes:
                if attr not in attrs:
                    attrs.append(attr)
        return tuple(attrs)

    @property
    def data_attributes(self) -> tuple[str]:
        attrs = []
        for data_set in self.data_sets:
            for attr in data_set.data_attributes:
                if attr not in attrs:
                    attrs.append(attr)
        return tuple(attrs)

    @property
    def dimensions(self) -> tuple[int]:
        dims = []
        for data_set in self.data_sets:
            for scen in data_set.scenarios:
                if scen.dim not in dims:
                    dims.append(scen.dim)
        return tuple(dims)

    @property
    def instances(self) -> tuple[int]:
        iids = []
        for data_set in self.data_sets:
            for scen in data_set.scenarios:
                for run in scen.runs:
                    if run.instance not in iids:
                        iids.append(run.instancem)
        return tuple(iids)
    
    @property
    def n_runs(self):
        return len(self.overview)

    @property
    def any(self):
        return len(self.data_sets) != 0

    def load(
        self,
        monotonic: bool = True,
        include_meta_data: bool = False,
        include_columns: list[str] = None,
        x_values: list[str] = None
    ) -> pl.DataFrame:
        if not self.any:
            return pl.DataFrame()

        data = []
        for data_set in self.data_sets:
            for scen in data_set.scenarios:
                df = scen.load(monotonic, data_set.function.maximization, x_values)
                data.append(df)

        data = pl.concat(data)

        if include_meta_data or include_columns is not None:
            if include_columns is None:
                include_columns = self.overview.columns

            for c in ("data_id", "run_id",):
                if c not in include_columns:
                    include_columns.append(c)
            
            data = self.overview.select(include_columns).join(
                data, on=("data_id", "run_id")
            )
        return data



import warnings
from typing import Iterable

import polars as pl
import numpy as np


def align_data(
    df: pl.DataFrame,
    evals: Iterable[int | float],
    group_cols: Iterable[str] = ("data_id",),
    x_col: str = "evaluations",
    y_col: str = "raw_y",
    output: str = "long",
    maximization: bool = False,
) -> pl.DataFrame:
    """Align data based on function evaluation counts

    Args:
        df (pl.DataFrame): DataFrame containing at minimum x, y and group columns specified in further parameters
        evals (Iterable[int]): list containing the function evaluation values at which to align
        group_cols (Iterable[str]): columns to use for grouping
        x_col (str, optional): function evaluation column Defaults to 'evaluations'.
        y_col (str, optional): function value column. Defaults to 'raw_y'.
        output (str, optional): whether to return a long or wide dataframe as output. Defaults to 'long'.
        maximization (bool, optional): whether the data comes from maximization or minimization. Defaults to False (minimization).

    Returns:
        pl.DataFrame: Alligned DataFrame
    """

    warnings.warn( "Turbo align is favoured over this function", DeprecationWarning)
    evals_df = pl.DataFrame({x_col: evals})

    def merge_asof_group(group):
        fval_col = x_col
        if x_col == "evaluations":
            fval_col = y_col

        if maximization:
            group = group.with_columns(group[fval_col].cum_max().alias(fval_col))
        else:
            group = group.with_columns(group[fval_col].cum_min().alias(fval_col))

        if x_col != "evaluations" and maximization:
            merged = evals_df.join_asof(
                group.sort(x_col), on=x_col, strategy="forward"
            ).fill_null(np.inf)
        else:
            merged = evals_df.join_asof(
                group.sort(x_col), on=x_col, strategy="backward"
            ).fill_null(np.inf)

        for col in group_cols:
            merged = merged.with_columns(pl.lit(group[col][0]).alias(col))
        return merged

    result_df = df.group_by(*group_cols).map_groups(merge_asof_group)

    if output == "long":
        return result_df

    pivot_df = result_df.pivot(index=x_col, columns=group_cols, values=y_col)
    return pivot_df


def turbo_align(
    df: pl.DataFrame,
    x_values: Iterable[int | float],
    x_col: str = "evaluations",
    y_col: str = "raw_y",
    output: str = "long",
    maximization: bool = False,
):
    """Align data based on function evaluation counts (fast

    Note:
        Assumes the data is monotonic!
        Assumes data_id is present -> i.e. data comes from manager

    Args:
        df (pl.DataFrame): DataFrame containing at minimum x, y and group columns specified in further parameters
        evals (Iterable[int]): list containing the function evaluation values at which to align
        x_col (str, optional): function evaluation column Defaults to 'evaluations'.
        y_col (str, optional): function value column. Defaults to 'raw_y'.
        output (str, optional): whether to return a long or wide dataframe as output. Defaults to 'long'.
        maximization (bool, optional): whether the data comes from maximization or minimization. Defaults to False (minimization).

    Returns:
        pl.DataFrame: Alligned DataFrame
    """

    data_ids = df["data_id"].unique()

    x_vals = pl.DataFrame(
        { 
            x_col: np.repeat(x_values, len(data_ids)),
            "data_id": np.tile(data_ids, len(x_values)),
        },
        schema={x_col: df[x_col].dtype, "data_id": df['data_id'].dtype},
    )
    df = df.sort([x_col, 'data_id'])
    
    if x_col != "evaluations" and maximization:
        result_df = x_vals.join_asof(
            df, by="data_id", on=x_col, strategy="forward"
        )
    else:
        result_df = x_vals.join_asof(
            df, by="data_id", on=x_col, strategy="backward"
        )
        
              
    if output == "long":
        return result_df

    pivot_df = result_df.pivot(index=x_col, columns=("data_id",), values=y_col)
    return pivot_df



import os
import unittest
import warnings

import polars as pl

from iohinspector import DataManager, turbo_align, plot_ecdf

from pprint import pprint

BASE_DIR = os.path.dirname(__file__)
DATA_DIR = os.path.realpath(os.path.join(BASE_DIR, "test_data"))


class TestManager(unittest.TestCase):

    def setUp(self):
        self.data_folders = [os.path.join(DATA_DIR, x) for x in sorted(os.listdir(DATA_DIR))]
        self.data_dir = self.data_folders[0]
        self.json_files = sorted(
            [
                fname
                for f in os.listdir(self.data_dir)
                if os.path.isfile((fname := os.path.join(self.data_dir, f)))
            ]
        )

    def test_add_json(self):
        manager = DataManager()
        manager.add_json(self.json_files[0])
        data = manager.data_sets[0]
        df = data.scenarios[0].load()
        self.assertTrue(isinstance(df, pl.DataFrame))
        self.assertEqual(max(df["run_id"]), 5)
        self.assertEqual(min(df["run_id"]), 1)
        self.assertEqual(len(df), 27)

    def test_load_twice(self):
        manager = DataManager()
        manager.add_json(self.json_files[0])
        self.assertEqual(len(manager.data_sets), 1)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            manager.add_json(self.json_files[0])
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, RuntimeWarning))

        self.assertEqual(len(manager.data_sets), 1)

    def test_add_folder(self):
        manager = DataManager()
        manager.add_folder(self.data_dir)
        self.assertEqual(len(manager.data_sets), 1)

    def test_select(self):
        manager = DataManager()
        manager.add_folders(self.data_folders)
        
        def assert_shape(df, n, m = 4):
            self.assertEqual(df.shape[1], m)
            self.assertEqual(len(df), n)
            self.assertEqual(max(df["run_id"]), 5)
            self.assertEqual(min(df["run_id"]), 1)
            self.assertTrue(selection.any)
            

        selection = manager.select(instances=[1], function_ids=[1])
        df = selection.load(monotonic=False)
        assert_shape(df, 84)
        df = selection.load(monotonic=True)
        assert_shape(df, 69)
        df = selection.load(monotonic=True, include_meta_data=True)
        assert_shape(df, 69, 13)

        selection = manager.select(function_ids=[0])
        self.assertFalse(selection.any)
        df = selection.load()
        self.assertEqual(len(df), 0)

        selection1 = manager.select(instances=[1], function_ids=[1])
        selection2 = manager.select(instances=[1], function_ids=[2])
        selection = selection1 + selection2
        df = selection.load()
        assert_shape(df, 125)

    def test_align(self):
        manager = DataManager()
        manager.add_folders(self.data_folders)
        
        selection = manager.select(function_ids=[1], algorithms = ['algorithm_A', 'algorithm_B'])
        df = selection.load(monotonic=True, include_meta_data=True)
        
        evals = [1, 5, 10, 20, 50, 100]
        df = turbo_align(df, evals)
        self.assertTrue(set(df['evaluations'].unique()) == set(evals))
        self.assertEqual(len(df['data_id'].unique()) * len(evals), df.shape[0])
        
    def test_plot_ecdf(self):
        manager = DataManager()
        manager.add_folders(self.data_folders)
        
        selection = manager.select(function_ids=[1], algorithms = ['algorithm_A', 'algorithm_B'])
        df = selection.load(monotonic=True, include_meta_data=True)
        
        dt = plot_ecdf(df)
        self.assertEqual(dt.shape, (66, 14))
        

    def test_select_on_data_id(self):
        manager = DataManager()
        manager.add_folders(self.data_folders)

        selection = manager.select(data_ids=[20, 21, 22])
        self.assertEqual(selection.n_runs, 3)
        
    def test_load_subset_columns(self):
        manager = DataManager()
        manager.add_folders(self.data_folders)
        selection = manager.select([1]).load(include_columns=["function_id"])
        self.assertListEqual(selection.columns, ["function_id", "data_id", "run_id", "evaluations", "raw_y"])


if __name__ == "__main__":
    pass



import os
import json
import warnings
from dataclasses import dataclass, field

import numpy as np
import polars as pl
from .align import turbo_align

METADATA_SCHEMA = [
    ("data_id", pl.UInt64),
    ("algorithm_name", pl.String),
    ("algorithm_info", pl.String),
    ("suite", pl.String),
    ("function_name", pl.String),
    ("function_id", pl.UInt16),
    ("dimension", pl.UInt16),
    ("instance", pl.UInt16),
    ("run_id", pl.UInt32),
    ("evals", pl.UInt64),
    ("best_y", pl.Float64),
]


def check_keys(data: dict, required_keys: list[str]):
    for key in required_keys:
        if key not in data:
            raise ValueError(
                f"data dict doesn't contain ioh format required key: {key}"
            )


def try_eval(value: str):
    try:
        return eval(value)
    except:
        return value


def get_polars_type(value):
    if isinstance(value, bool):
        return pl.Boolean
    if isinstance(value, int):
        return pl.Int64
    if isinstance(value, float):
        return pl.Float64
    if isinstance(value, str):
        return pl.String

    warnings.warn(f"{type(value)} is not mapped to polars dtype", UserWarning)
    return pl.Object


@dataclass
class Function:
    id: int
    name: str
    maximization: bool


@dataclass
class Algorithm:
    name: str
    info: str


@dataclass
class Solution:
    evals: int
    x: np.ndarray = field(repr=None)
    y: float


@dataclass
class Run:
    data_id: int
    id: int
    instance: int
    evals: int
    best: Solution

    __lookup__ = {}
    __current_id__ = 1

    @staticmethod
    def hash(key: str):
        if value := Run.__lookup__.get(key):
            return value
        Run.__lookup__[key] = Run.__current_id__
        Run.__current_id__ += 1
        return Run.__lookup__[key]


@dataclass
class Scenario:
    dimension: int
    data_file: str
    runs: list[Run]

    @staticmethod
    def from_dict(data: dict, dirname: str):
        """Constructs a Scenario object from a dictionary
        (output of json.load from ioh compatible file)
        """

        required_keys = (
            "dimension",
            "path",
            "runs",
        )
        check_keys(data, required_keys)

        data["path"] = os.path.join(dirname, data["path"])
        if not os.path.isfile(data["path"]):
            raise FileNotFoundError(f"{data['path']} is not found")

        return Scenario(
            data["dimension"],
            data["path"],
            [
                Run(
                    Run.hash(f"{data['path']}_{run_id}"),
                    run_id,
                    run["instance"],
                    run["evals"],
                    best=Solution(**run["best"]),
                )
                for run_id, run in enumerate(data["runs"], 1)
            ],
        )

    def load(self, monotonic=False, maximize=True, x_values = None) -> pl.DataFrame:
        """Loads the data file stored at self.data_file to a pd.DataFrame"""

        with open(self.data_file) as f:
            header = next(f).strip().split()

        key_lookup = dict([(r.id, r.data_id) for r in self.runs])
        dt = (
            pl.scan_csv(
                self.data_file,
                separator=" ",
                decimal_comma=True,
                schema={header[0]: pl.Float64, **dict.fromkeys(header[1:], pl.Float64)},
                ignore_errors=True,
                
            )
            .with_columns(
                pl.col("evaluations").cast(pl.UInt64),
                run_id=(pl.col("evaluations") == 1).cum_sum(),
            )
            .drop_nulls()
            .filter(pl.col("run_id").is_in([r.id for r in self.runs]))
            .with_columns(
                data_id=pl.col("run_id").map_elements(
                    key_lookup.__getitem__, return_dtype=pl.UInt64
                )
            )
        )

        if monotonic or x_values is not None:
            if maximize:
                dt = dt.with_columns(pl.col("raw_y").cum_max().over("run_id"))
            else:
                dt = dt.with_columns(pl.col("raw_y").cum_min().over("run_id"))

            dt = dt.filter(pl.col("raw_y").diff().fill_null(1.0).abs() > 0.0)
            
            
        dt = dt.collect()
        
        if x_values is not None:
            dt = turbo_align(dt, x_values)                        
        
        return dt


@dataclass
class Dataset:
    file: str
    version: str
    suite: str
    function: Function
    algorithm: Algorithm
    experiment_attributes: list[tuple[str, str]]
    data_attributes: list[str]
    scenarios: list[Scenario]

    @staticmethod
    def from_json(json_file: str):
        """Construct a dataset object from a json file"""

        if not os.path.isfile(json_file):
            raise FileNotFoundError(f"{json_file} not found")

        with open(json_file) as f:
            data = json.load(f)
            return Dataset.from_dict(data, json_file)

    @property
    def overview(self) -> pl.DataFrame:
        meta_data = [
            self.algorithm.name,
            self.algorithm.info,
            self.suite,
            self.function.name,
            self.function.id,
        ]
        if self.experiment_attributes:
            exattr_names, exattr_values = zip(*self.experiment_attributes)
            exattr_values = list(map(try_eval, exattr_values))
            exattr_schema = [
                (name, get_polars_type(value))
                for name, value in zip(exattr_names, exattr_values)
            ]
        else:
            exattr_values = []
            exattr_schema = []

        records = []
        for scen in self.scenarios:
            for run in scen.runs:
                records.append(
                    [run.data_id]
                    + meta_data
                    + [scen.dimension, run.instance, run.id, run.evals, run.best.y]
                    + exattr_values
                )
        return pl.DataFrame(records, schema=METADATA_SCHEMA + exattr_schema, orient="row") 

    @staticmethod
    def from_dict(data: dict, filepath: str):
        """Constructs a Dataset object from a dictionary
        (output of json.load from ioh compatible file)
        """

        required_keys = (
            "version",
            "suite",
            "function_id",
            "function_name",
            "maximization",
            "algorithm",
            # "experiment_attributes",
            "attributes",
            "scenarios",
        )
        check_keys(data, required_keys)

        if "experiment_attributes" in data:
            experiment_attributes = [tuple(x.items())[0] for x in data["experiment_attributes"]]
        else:
            experiment_attributes = None

        return Dataset(
            filepath,
            data["version"],
            data["suite"],
            Function(data["function_id"], data["function_name"], data["maximization"]),
            Algorithm(
                data["algorithm"]["name"],
                data["algorithm"]["info"],
            ),
            experiment_attributes,
            data["attributes"],
            [
                Scenario.from_dict(scen, os.path.dirname(filepath))
                for scen in data["scenarios"]
            ],
        )



from typing import Iterable

import numpy as np
import polars as pl

from moocore import (
    hypervolume,
    igd_plus,
    epsilon_additive,
    epsilon_mult,
    filter_dominated,
)


def get_reference_set(
    data: pl.DataFrame, obj_cols: Iterable[str], max_size: int = 1000
) -> np.ndarray:
    """Get a (subsampled) reference set from a set of data

    Args:
        data (pl.DataFrame): The dataframe from which to extract the objective values
        obj_cols (Iterable[str]): The names of the objectives in 'data'
        max_size (int, optional): The maximum number of points in the reference set. Defaults to 1000.

    Returns:
        np.ndarray: The filtered reference set
    """
    obj_vals = np.array(data[obj_cols])
    ref_set = filter_dominated(obj_vals)
    if len(ref_set) < max_size:
        return ref_set
    return ref_set[np.random.choice(len(ref_set), size=max_size)]


def _tchebycheff(weight_vec, ideal_point, point):
    tch_val = -np.inf
    for w, z, f in zip(weight_vec, ideal_point, point):
        tch_val = max(tch_val, w * abs(f - z))
    return tch_val


def _r2(weight_vec_set, ideal_point, point_set):
    sum_tch = 0
    for w in weight_vec_set:
        min_tch = np.inf
        for p in point_set:
            tch = _tchebycheff(w, ideal_point, p)
            min_tch = min(min_tch, tch)
        sum_tch += min_tch
    return sum_tch / len(weight_vec_set)


class NonDominated:
    def __call__(self, group: pl.DataFrame, objective_columns: Iterable):
        objectives = np.array(group[objective_columns])
        is_efficient = np.ones(objectives.shape[0], dtype=bool)
        for i, c in enumerate(objectives[1:]):
            if is_efficient[i + 1]:
                is_efficient[i + 1 :][is_efficient[i + 1 :]] = np.any(
                    objectives[i + 1 :][is_efficient[i + 1 :]] < c, axis=1
                )
                is_efficient[i + 1] = True
        group = group.with_columns(pl.Series(name="nondominated", values=is_efficient))
        return group


class HyperVolume:
    def __init__(self, reference_point: np.ndarray):
        """Function to calculate the Hypervolume metric over time. Used as an input to the 'add_indicator' function.

        Args:
            reference_set (np.ndarray): The reference point for the HV calculation.
        """
        self.reference_point = reference_point

    @property
    def var_name(self):
        return "HyperVolume"

    @property
    def minimize(self):
        return False

    def __call__(
        self, group: pl.DataFrame, objective_columns: Iterable, evals: Iterable[int]
    ) -> pl.DataFrame:
        """
        Args:
            group (pl.DataFrame): The DataFrame on which the indicator will be added (should be 1 optimization run only)
            objective_columns (Iterable): Which columns are the objectives
            evals (Iterable[int]): At which evaluations the operation should be performed.
            Note that using more evaluations will make the code slower.

        Returns:
            pl.DataFrame: a new DataFrame with columns of 'evals' and corresponding IGD+
        """
        obj_vals = np.clip(
            np.array(group[objective_columns]), None, self.reference_point
        )
        evals_dt = group["evaluations"]
        hvs = [
            hypervolume(obj_vals[: (evals_dt <= eval).sum()], ref=self.reference_point)
            for eval in evals
        ]
        return (
            pl.DataFrame(
                [
                    pl.Series(name="evaluations", values=evals, dtype=pl.UInt64),
                    pl.Series(name="HyperVolume", values=hvs),
                ]
            )
            .join_asof(group.sort("evaluations"), on="evaluations", strategy="backward")
            .fill_null(np.inf)
            .drop(objective_columns)
        )


class Epsilon:
    def __init__(self, reference_point: np.ndarray, version: str = "additive"):
        """Function to calculate the Epsilon indicator over time. Used as an input to the 'add_indicator' function.

        Args:
            reference_set (np.ndarray): The reference point for the HV calculation.
            version (str): Whether to use the additive or multiplicative version of this indicator.
        """
        self.reference_point = reference_point
        if version == "additive":
            self.indicator = epsilon_additive
            self._var_name = "Epsilon_Additive"
        else:
            self.indicator = epsilon_mult
            self._var_name = "Epsilon_Mult"

    @property
    def var_name(self):
        return self._var_name

    @property
    def minimize(self):
        return True

    def __call__(
        self, group: pl.DataFrame, objective_columns: Iterable, evals: Iterable[int]
    ) -> pl.DataFrame:
        """
        Args:
            group (pl.DataFrame): The DataFrame on which the indicator will be added (should be 1 optimization run only)
            objective_columns (Iterable): Which columns are the objectives
            evals (Iterable[int]): At which evaluations the operation should be performed.
            Note that using more evaluations will make the code slower.

        Returns:
            pl.DataFrame: a new DataFrame with columns of 'evals' and corresponding IGD+
        """
        obj_vals = np.clip(
            np.array(group[objective_columns]), None, self.reference_point
        )
        evals_dt = group["evaluations"]
        hvs = [
            self.indicator(
                filter_dominated(obj_vals[: (evals_dt <= eval).sum()]),
                ref=self.reference_point,
            )
            for eval in evals
        ]
        return (
            pl.DataFrame(
                [
                    pl.Series(name="evaluations", values=evals, dtype=pl.UInt64),
                    pl.Series(name=self._var_name, values=hvs),
                ]
            )
            .join_asof(group.sort("evaluations"), on="evaluations", strategy="backward")
            .fill_null(np.inf)
            .drop(objective_columns)
        )


class IGDPlus:
    def __init__(self, reference_set: np.ndarray):
        """Function to calculate the IGD+ metric over time. Used as an input to the 'add_indicator' function.

        Args:
            reference_set (np.ndarray): The reference set for the IGD+ calculation. Note that larger sets make
            the calculation slower
        """
        self.reference_set = reference_set

    @property
    def minimize(self):
        return True

    @property
    def var_name(self):
        return "IGD+"

    def __call__(
        self, group: pl.DataFrame, objective_columns: Iterable, evals: Iterable[int]
    ) -> pl.DataFrame:
        """

        Args:
            group (pl.DataFrame): The DataFrame on which the indicator will be added (should be 1 optimization run only)
            objective_columns (Iterable): Which columns are the objectives
            evals (Iterable[int]): At which evaluations the operation should be performed.
            Note that using more evaluations will make the code slower.

        Returns:
            pl.DataFrame: a new DataFrame with columns of 'evals' and corresponding IGD+
        """
        obj_vals = np.array(group[objective_columns])
        evals_dt = group["evaluations"]
        igds = [
            igd_plus(
                filter_dominated(obj_vals[: (evals_dt <= eval).sum()]),
                ref=self.reference_set,
            )
            for eval in evals
        ]
        return (
            pl.DataFrame(
                [
                    pl.Series(name="evaluations", values=evals, dtype=pl.UInt64),
                    pl.Series(name="IGD+", values=igds),
                ]
            )
            .join_asof(group.sort("evaluations"), on="evaluations", strategy="backward")
            .fill_null(np.inf)
            .drop(objective_columns)
        )

try:
    from pymoo.util.ref_dirs import get_reference_directions
    class R2:
        def __init__(self, n_ref_dirs: int, ideal_point: np.ndarray):
            """Function to calculate the R2 indicator over time. Used as an input to the 'add_indicator' function.

            Args:
                n_ref_dirs (int): How many reference directions to use. Reference directions are generated based on pymoo's 'energy' method.
                ideal_point (np.ndarray): The ideal point for the R2 calculations
            """
            self.ref_dirs = get_reference_directions("energy", len(ideal_point), n_ref_dirs)
            self.ideal_point = ideal_point

        @property
        def var_name(self):
            return "R2"

        @property
        def minimize(self):
            return True

        def __call__(
            self, group: pl.DataFrame, objective_columns: Iterable, evals: Iterable[int]
        ) -> pl.DataFrame:
            """

            Args:
                group (pl.DataFrame): The DataFrame on which the indicator will be added (should be 1 optimization run only)
                objective_columns (Iterable): Which columns are the objectives
                evals (Iterable[int]): At which evaluations the operation should be performed.
                Note that using more evaluations will make the code slower.

            Returns:
                pl.DataFrame: a new DataFrame with columns of 'evals' and corresponding IGD+
            """
            obj_vals = np.array(group[objective_columns])
            evals_dt = group["evaluations"]
            igds = [
                _r2(
                    self.ref_dirs,
                    self.ideal_point,
                    filter_dominated(obj_vals[: (evals_dt <= eval).sum()]),
                )
                for eval in evals
            ]
            return (
                pl.DataFrame(
                    [
                        pl.Series(name="evaluations", values=evals, dtype=pl.UInt64),
                        pl.Series(name=self.var_name, values=igds),
                    ]
                )
                .join_asof(group.sort("evaluations"), on="evaluations", strategy="backward")
                .fill_null(np.inf)
                .drop(objective_columns)
            )

except ImportError:
    class R2:
        def __init__(self, *args, **kwargs):
            import warnings
            warnings.warn("R2 indicator is N/A without pymoo installed")


from typing import Iterable

import polars as pl
import numpy as np
from moocore import is_nondominated


class NonDominated:
    def __call__(self, group: pl.DataFrame, objective_columns: Iterable):
        objectives = np.array(group[objective_columns])
        return group.with_columns(
            pl.Series(name="final_nondominated", values=is_nondominated(objectives))
        )



from functools import partial
from warnings import warn
from typing import Iterable, Callable, Optional

import polars as pl
import numpy as np
import pandas as pd
from skelo.model.elo import EloEstimator

from .align import align_data




def get_sequence(
    min: float,
    max: float,
    len: float,
    scale_log: bool = False,
    cast_to_int: bool = False,
) -> np.ndarray:
    """Create sequence of points, used for subselecting targets / budgets for allignment and data processing

    Args:
        min (float): Starting point of the range
        max (float): Final point of the range
        len (float): Number of steps
        scale_log (bool): Whether values should be scaled logarithmically. Defaults to False
        version (str, optional): Whether the value should be casted to integers (e.g. in case of budget) or not. Defaults to False.

    Returns:
        np.ndarray: Array of evenly spaced values
    """
    transform = lambda x: x
    if scale_log:
        assert min > 0
        min = np.log10(min)
        max = np.log10(max)
        transform = lambda x: 10**x
    values = transform(
        np.arange(
            min,
            max + (max - min) / (2 * (len - 1)),
            (max - min) / (len - 1),
            dtype=float,
        )
    )
    if cast_to_int:
        return np.unique(np.array(values, dtype=int))
    return np.unique(values)


def _geometric_mean(series: pl.Series) -> float:
    """Helper function for polars: geometric mean"""
    return np.exp(np.log(series).mean())


def aggegate_convergence(
    data: pl.DataFrame,
    evaluation_variable: str = "evaluations",
    fval_variable: str = "raw_y",
    free_variables: Iterable[str] = ["algorithm_name"],
    x_min: int = None,
    x_max: int = None,
    custom_op: Callable[[pl.Series], float] = None,
    maximization: bool = False,
    return_as_pandas: bool = True,
):
    """Function to aggregate performance on a fixed-budget perspective

    Args:
        data (pl.DataFrame): The data object to use for getting the performance. Note that the fval, evaluation and free variables as defined in
        this object determine the axes of the final performance (most data will have 'raw_y', 'evaluations' and ['algId'] as defaults)
        evaluation_variable (str, optional): Column name for evaluation number. Defaults to "evaluations".
        fval_variable (str, optional): Column name for function value. Defaults to "raw_y".
        free_variables (Iterable[str], optional): Column name for free variables (variables over which performance should not be aggregated). Defaults to ["algorithm_name"].
        x_min (int, optional): Minimum evaulation value to use. Defaults to None (minimum present in data).
        x_max (int, optional): Maximum evaulation value to use. Defaults to None (maximum present in data).
        custom_op (Callable[[pl.Series], float], optional): Custom aggregation method for performance values. Defaults to None.
        maximization (bool, optional): Whether performance metric is being maximized or not. Defaults to False.
        return_as_pandas (bool, optional): Whether the data should be returned as Pandas (True) or Polars (False) object. Defaults to True.

    Returns:
        DataFrame: Depending on 'return_as_pandas', a pandas or polars DataFrame with the aggregated performance values
    """

    # Getting alligned data (to check if e.g. limits should be args for this function)
    if x_min is None:
        x_min = data[evaluation_variable].min()
    if x_max is None:
        x_max = data[evaluation_variable].max()
    x_values = get_sequence(x_min, x_max, 50, scale_log=True, cast_to_int=True)
    group_variables = free_variables + [evaluation_variable]
    data_aligned = align_data(
        data.cast({evaluation_variable: pl.Int64}),
        x_values,
        group_cols=["data_id"] + free_variables,
        x_col=evaluation_variable,
        y_col=fval_variable,
        maximization=maximization,
    )

    aggregations = [
        pl.mean(fval_variable).alias("mean"),
        pl.min(fval_variable).alias("min"),
        pl.max(fval_variable).alias("max"),
        pl.median(fval_variable).alias("median"),
        pl.std(fval_variable).alias("std"),
        pl.col(fval_variable)
        .map_elements(lambda s: _geometric_mean(s), return_dtype=pl.Float64)
        .alias("geometric_mean"),
    ]

    if custom_op is not None:
        aggregations.append(
            pl.col(evaluation_variable)
            .map_elements(lambda s: custom_op(s), return_dtype=pl.Float64)
            .alias(custom_op.__name__)
        )
    dt_plot = data_aligned.group_by(*group_variables).agg(aggregations)
    if return_as_pandas:
        return dt_plot.sort(evaluation_variable).to_pandas()
    return dt_plot.sort(evaluation_variable)


def transform_fval(
    data: pl.DataFrame,
    lb: float = 1e-8,
    ub: float = 1e8,
    scale_log: bool = True,
    maximization: bool = False,
    fval_col: str = "raw_y",
):
    """Helper function to transform function values (min-max normalization based on provided bounds and scaling)

    Args:
        data (pl.DataFrame): The data object to use for getting the performance.
        lb (float, optional): Lower bound for scaling of function values. If None, it is the max value found in data. Defaults to 1e-8.
        ub (float, optional): Upper bound for scaling of function values. If None, it is the max value found in data. Defaults to 1e8.
        scale_log (bool, optional): Whether function values should be log-scaled before scaling. Defaults to True.
        maximization (bool, optional): Whether function values is being maximized. Defaults to False.
        fval_col (str, optional): Which column in data to use. Defaults to "raw_y".

    Returns:
        _type_: a copy of the original data with a new column 'eaf' with the scaled function values (which is always to be maximized)
    """
    if ub == None:
        ub = data[fval_col].max()
    if lb == None:
        lb = data[fval_col].min()
        if lb <= 0 and scale_log:
            lb = 1e-8
            warnings.warn(
                "If using logarithmic scaling, lb should be set to prevent errors in log-calculation. Lb is being overwritten to 1e-8 to avoid this."
            )
    if scale_log:
        lb = np.log10(lb)
        ub = np.log10(ub)
        res = data.with_columns(
            ((pl.col(fval_col).log10() - lb) / (ub - lb)).clip(0, 1).alias("eaf")
        )
    else:
        res = data.with_columns(
            ((pl.col(fval_col) - lb) / (ub - lb)).clip(0, 1).alias("eaf")
        )
    if maximization:
        return res
    return res.with_columns((1 - pl.col("eaf")).alias("eaf"))


def _aocc(group: pl.DataFrame, max_budget: int, fval_col: str = "eaf"):
    group = group.cast({"evaluations": pl.Int64}).filter(
        pl.col("evaluations") <= max_budget
    )
    new_row = pl.DataFrame(
        {
            "evaluations": [0, max_budget],
            fval_col: [group[fval_col].min(), group[fval_col].max()],
        }
    )
    group = (
        pl.concat([group, new_row], how="diagonal")
        .sort("evaluations")
        .fill_null(strategy="forward")
        .fill_null(strategy="backward")
    )
    return group.with_columns(
        (
            (
                pl.col("evaluations").diff(n=1, null_behavior="ignore")
                * (pl.col(fval_col).shift(1))
            )
            / max_budget
        ).alias("aocc_contribution")
    )


def get_aocc(
    data: pl.DataFrame,
    max_budget: int,
    fval_col: str = "eaf",
    group_cols: Iterable[str] = ["function_name", "algorithm_name"],
):
    """Helper function for AOCC calculations

    Args:
        data (pl.DataFrame): The data object to use for getting the performance.
        max_budget (int): Maxium value of evaluations to use
        fval_col (str, optional): Which data column specifies the performance value. Defaults to "eaf".
        group_cols (Iterable[str], optional): Which columns to NOT aggregate over. Defaults to ["function_name", "algorithm_name"].

    Returns:
        pl.DataFrame: a polars dataframe with the area under the EAF (=area over convergence curve)
    """
    aocc_contribs = data.group_by(*["data_id"]).map_groups(
        partial(_aocc, max_budget=max_budget, fval_col=fval_col)
    )
    aoccs = aocc_contribs.group_by(["data_id"] + group_cols).agg(
        pl.col("aocc_contribution").sum()
    )
    return aoccs.group_by(group_cols).agg(
        pl.col("aocc_contribution").mean().alias("AOCC")
    )


def get_tournament_ratings(
    data: pl.DataFrame,
    alg_vars: Iterable[str] = ["algorithm_name"],
    fid_vars: Iterable[str] = ["function_name"],
    perf_var: str = "raw_y",
    nrounds: int = 25,
    maximization: bool = False,
):
    """Method to calculate ratings of a set of algorithm on a set of problems.
    Calculated based on nrounds of competition, where in each round all algorithms face all others (pairwise) on every function.
    For each round, a sampled performance value is taken from the data and used to determine the winner.
    This function uses the ELO rating scheme, as opposed to the Glicko2 scheme used in the IOHanalyzer. Deviations are estimated based on the last 5% of rounds.

    Args:
        data (pl.DataFrame): The data object to use for getting the performance.
        alg_vars (Iterable[str], optional): Which variables specific the algortihms which will compete. Defaults to ["algorithm_name"].
        fid_vars (Iterable[str], optional): Which variables denote the problems on which will be competed. Defaults to ["function_name"].
        perf_var (str, optional): Which variable corresponds to the performance. Defaults to "raw_y".
        nrounds (int, optional): How many round should be played. Defaults to 25.
        maximization (bool, optional): Whether the performance metric is being maximized. Defaults to False.

    Returns:
        pd.DataFrame: Pandas dataframe with rating, deviation and volatility for each 'alg_vars' combination
    """
    fids = data[fid_vars].unique()
    aligned_comps = data.pivot(
        index=alg_vars,
        columns=fid_vars,
        values=perf_var,
        aggregate_function=pl.element(),
    )
    players = aligned_comps[alg_vars]
    n_players = players.shape[0]
    comp_arr = np.array(aligned_comps[aligned_comps.columns[len(alg_vars) :]])

    rng = np.random.default_rng()
    fids = [i for i in range(len(fids))]
    lplayers = [i for i in range(n_players)]
    records = []
    for r in range(nrounds):
        for fid in fids:
            for p1 in lplayers:
                for p2 in lplayers:
                    if p1 == p2:
                        continue
                    s1 = rng.choice(comp_arr[p1][fid], 1)[0]
                    s2 = rng.choice(comp_arr[p2][fid], 1)[0]
                    if s1 == s2:
                        won = 0.5
                    else:
                        won = abs(float(maximization) - float(s1 < s2))

                    records.append([r, p1, p2, won])
                    
    dt_comp = pd.DataFrame.from_records(
        records, columns=["round", "p1", "p2", "outcome"]
    )
    dt_comp = dt_comp.sample(frac=1).sort_values("round")
    model = EloEstimator(key1_field="p1", key2_field="p2", timestamp_field="round").fit(
        dt_comp, dt_comp["outcome"]
    )
    model_dt = model.rating_model.to_frame()
    ratings = np.array(model_dt[np.isnan(model_dt["valid_to"])]["rating"])
    deviations = (
        model_dt.query(f"valid_from >= {nrounds * 0.95}").groupby("key")["rating"].std()
    )
    rating_dt_elo = pd.DataFrame(
        [
            ratings,
            deviations,
            *players[players.columns],
        ]
    ).transpose()
    rating_dt_elo.columns = ["Rating", "Deviation", *players.columns]
    return rating_dt_elo


def aggegate_running_time(
    data: pl.DataFrame,
    evaluation_variable: str = "evaluations",
    fval_variable: str = "raw_y",
    free_variables: Iterable[str] = ["algorithm_name"],
    f_min: float = None,
    f_max: float = None,
    scale_flog: bool = True,
    max_budget: int = None,
    maximization: bool = False,
    custom_op: Callable[[pl.Series], float] = None,
    return_as_pandas: bool = True,
):
    """Function to aggregate performance on a fixed-target perspective

    Args:
        data (pl.DataFrame): The data object to use for getting the performance. Note that the fval, evaluation and free variables as defined in
        this object determine the axes of the final performance (most data will have 'raw_y', 'evaluations' and ['algId'] as defaults)
        evaluation_variable (str, optional): Column name for evaluation number. Defaults to "evaluations".
        fval_variable (str, optional): Column name for function value. Defaults to "raw_y".
        free_variables (Iterable[str], optional): Column name for free variables (variables over which performance should not be aggregated). Defaults to ["algorithm_name"].
        f_min (int, optional): Minimum function value to use. Defaults to None (minimum present in data).
        f_max (int, optional): Maximum function value to use. Defaults to None (maximum present in data).
        scale_flog (bool): Whether or not function values should be scaled logarithmically for the x-axis. Defaults to True.
        max_budget: If present, what budget value should be the maximum considered. Defaults to None.
        custom_op (Callable[[pl.Series], float], optional): Custom aggregation method for performance values. Defaults to None.
        maximization (bool, optional): Whether performance metric is being maximized or not. Defaults to False.
        return_as_pandas (bool, optional): Whether the data should be returned as Pandas (True) or Polars (False) object. Defaults to True.

    Returns:
        DataFrame: Depending on 'return_as_pandas', a pandas or polars DataFrame with the aggregated performance values
    """

    # Getting alligned data (to check if e.g. limits should be args for this function)
    if f_min is None:
        f_min = data[fval_variable].min()
    if f_max is None:
        f_max = data[fval_variable].max()
    f_values = get_sequence(f_min, f_max, 50, scale_log=scale_flog)
    group_variables = free_variables + [fval_variable]
    data_aligned = align_data(
        data,
        f_values,
        group_cols=["data_id"] + free_variables,
        x_col=fval_variable,
        y_col=evaluation_variable,
        maximization=maximization,
    )
    if max_budget is None:
        max_budget = data[evaluation_variable].max()

    aggregations = [
        pl.col(evaluation_variable).replace(np.inf, max_budget).mean().alias("mean"),
        # pl.mean(evaluation_variable).alias("mean"),
        pl.col(evaluation_variable).replace(np.inf, max_budget).min().alias("min"),
        pl.col(evaluation_variable).replace(np.inf, max_budget).max().alias("max"),
        pl.col(evaluation_variable)
        .replace(np.inf, max_budget)
        .median()
        .alias("median"),
        pl.col(evaluation_variable).replace(np.inf, max_budget).std().alias("std"),
        pl.col(evaluation_variable).is_finite().mean().alias("success_ratio"),
        pl.col(evaluation_variable).is_finite().sum().alias("success_count"),
        (
            pl.col(evaluation_variable).replace(np.inf, max_budget).sum()
            / pl.col(evaluation_variable).is_finite().sum()
        ).alias("ERT"),
        (
            pl.col(evaluation_variable).replace(np.inf, max_budget * 10).sum()
            / pl.col(evaluation_variable).count()
        ).alias("PAR-10"),
    ]

    if custom_op is not None:
        aggregations.append(
            pl.col(evaluation_variable)
            .apply(lambda s: custom_op(s))
            .alias(custom_op.__name__)
        )
    dt_plot = data_aligned.group_by(*group_variables).agg(aggregations)
    if return_as_pandas:
        return dt_plot.sort(fval_variable).to_pandas()
    return dt_plot.sort(fval_variable)


def add_normalized_objectives(
    data: pl.DataFrame, obj_cols: Iterable[str], max_vals: Optional[pl.DataFrame] = None, min_vals: Optional[pl.DataFrame] = None
):
    """Add new normalized columns to provided dataframe based on the provided objective columns

    Args:
        data (pl.DataFrame): The original dataframe
        obj_cols (Iterable[str]): The names of each objective column
        max_vals (Optional[pl.DataFrame]): If provided, these values will be used as the maxima instead of the values found in `data`
        min_vals (Optional[pl.DataFrame]): If provided, these values will be used as the minima instead of the values found in `data`

    Returns:
        _type_: The original `data` DataFrame with a new column 'objI' added for each objective, for I=1...len(obj_cols)
    """
    if type(max_vals) == pl.DataFrame:
        data_max = [max_vals[colname].max() for colname in obj_cols]
    else:
        data_max = [data[colname].max() for colname in obj_cols]
    if type(min_vals) == pl.DataFrame:
        data_min = [min_vals[colname].min() for colname in obj_cols]
    else:
        data_min = [data[colname].min() for colname in obj_cols]
    return data.with_columns(
        [
            ((data[colname] - data_min[idx]) / (data_max[idx] - data_min[idx])).alias(f"obj{idx + 1}")
            for idx, colname in enumerate(obj_cols)
        ]
    )


def _get_nodeidx(xloc, yval, nodes, epsilon):
    if len(nodes) == 0:
        return -1
    candidates = nodes[np.isclose(nodes["y"], yval, atol=epsilon)]
    if len(candidates) == 0:
        return -1
    idxs = np.all(
        np.isclose(np.array(candidates)[:, : len(xloc)], xloc, atol=epsilon), axis=1
    )
    if any(idxs):
        return candidates[idxs].index[0]
    return -1


def get_attractor_network(
    data,
    coord_vars=["x1", "x2"],
    fval_var: str = "raw_y",
    eval_var: str = "evaluations",
    maximization: bool = False,
    beta=40,
    epsilon=0.0001,
    eval_max=None,
):
    """Create an attractor network from the provided data

    Args:
        data (pl.DataFrame): The original dataframe, should contain the performance and position information
        coord_vars (Iterable[str], optional): Which columns correspond to position information. Defaults to ['x1', 'x2'].
        fval_var (str, optional): Which column corresponds to performance. Defaults to 'raw_y'.
        eval_var (str, optional): Which column corresponds to evaluations. Defaults to 'evaluations'.
        maximization (bool, optional): Whether fval_var is to be maximized. Defaults to False.
        beta (int, optional): Minimum stagnation lenght. Defaults to 40.
        epsilon (float, optional): Radius below which positions should be considered identical in the network. Defaults to 0.0001.
        eval_max (int, optional): Maximum evaluation number. Defaults to the maximum of eval_var if None.
    Returns:
        pd.DataFrame, pd.DataFrame: two dataframes containing the nodes and edges of the network respectively.
    """

    running_idx = 0
    running_edgeidx = 0
    nodes = pd.DataFrame(columns=[*coord_vars, "y", "count", "evals"])
    edges = pd.DataFrame(columns=["start", "end", "count", "stag_length_avg"])
    if eval_max is None:
        eval_max = max(data[eval_var])

    for run_id in data["data_id"].unique():
        dt_group = data.filter(
            pl.col("data_id") == run_id, pl.col(eval_var) <= eval_max
        )
        if maximization:
            ys = np.maximum.accumulate(np.array(dt_group[fval_var]))
        else:
            ys = np.minimum.accumulate(np.array(dt_group[fval_var]))
        xs = np.array(dt_group[coord_vars])

        stopping_points = np.where(np.abs(np.diff(ys, prepend=np.inf)) > 0)[0]
        evals = np.array(dt_group[eval_var])

        stagnation_lengths = np.diff(evals[stopping_points], append=eval_max)
        edge_lengths = stagnation_lengths[stagnation_lengths > beta]
        real_idxs = [stopping_points[i] for i in np.where(stagnation_lengths > beta)[0]]

        xloc = xs[real_idxs[0]]
        yval = ys[real_idxs[0]]
        nodeidx = _get_nodeidx(xloc, yval, nodes, epsilon)
        if nodeidx == -1:
            nodes.loc[running_idx] = [*xloc, yval, 1, evals[real_idxs[0]]]
            node1 = running_idx
            running_idx += 1
        else:
            nodes.loc[nodeidx, "evals"] += evals[real_idxs[0]]
            nodes.loc[nodeidx, "count"] += 1
            node1 = nodeidx

        if len(real_idxs) == 1:
            continue

        for i in range(len(real_idxs) - 1):
            xloc = xs[real_idxs[i + 1]]
            yval = ys[real_idxs[i + 1]]
            nodeidx = _get_nodeidx(xloc, yval, nodes, epsilon)
            if nodeidx == -1:
                nodes.loc[running_idx] = [*xloc, yval, 1, evals[real_idxs[i + 1]]]
                node2 = running_idx
                running_idx += 1
            else:
                nodes.loc[nodeidx, "evals"] += evals[real_idxs[i + 1]]
                nodes.loc[nodeidx, "count"] += 1
                node2 = nodeidx

            edgelen = edge_lengths[i]
            edge_idxs = edges.query(f"start == {node1} & end == {node2}").index
            if len(edge_idxs) == 0:
                edges.loc[running_edgeidx] = [node1, node2, 1, edgelen]
                running_edgeidx += 1
            else:
                curr_count = edges.loc[edge_idxs[0]]["count"]
                curr_len = edges.loc[edge_idxs[0]]["stag_length_avg"]
                edges.loc[edge_idxs[0], "stag_length_avg"] = (
                    curr_len * curr_count + edgelen
                ) / (curr_count + 1)
                edges.loc[edge_idxs[0], "count"] += 1
            node1 = node2
    return nodes, edges


def get_data_ecdf(
    data,
    fval_var: str = "raw_y",
    eval_var: str = "evaluations",
    free_vars: Iterable[str] = ["algorithm_name"],
    maximization: bool = False,
    x_values: Iterable[int] = None,
    x_min: int = None,
    x_max: int = None,
    scale_xlog: bool = True,
    y_min: int = None,
    y_max: int = None,
    scale_ylog: bool = True,
):
    """Function to plot empirical cumulative distribution function (Based on EAF)

    Args:
        data (pl.DataFrame): The DataFrame which contains the full performance trajectory. Should be generated from a DataManager.
        eval_var (str, optional): Column in 'data' which corresponds to the number of evaluations. Defaults to "evaluations".
        fval_var (str, optional): Column in 'data' which corresponds to the performance measure. Defaults to "raw_y".
        free_vars (Iterable[str], optional): Columns in 'data' which correspond to groups over which data should not be aggregated. Defaults to ["algorithm_name"].
        maximization (bool, optional): Boolean indicating whether the 'fval_var' is being maximized. Defaults to False.
        measures (Iterable[str], optional): List of measures which should be used in the plot. Valid options are 'geometric_mean', 'mean', 'median', 'min', 'max'. Defaults to ['geometric_mean'].
        x_values (Iterable[int], optional): List of x-values at which to get the ECDF data. If not provided, the x_min, x_max and scale_xlog arguments will be used to sample these points.
        scale_xlog (bool, optional): Should the x-samples be log-scaled. Defaults to True.
        x_min (float, optional): Minimum value to use for the 'eval_var', if not present the min of that column will be used. Defaults to None.
        x_max (float, optional): Maximum value to use for the 'eval_var', if not present the max of that column will be used. Defaults to None.
        scale_ylog (bool, optional): Should the y-values be log-scaled before normalization. Defaults to True.
        y_min (float, optional): Minimum value to use for the 'fval_var', if not present the min of that column will be used. Defaults to None.
        y_max (float, optional): Maximum value to use for the 'fval_var', if not present the max of that column will be used. Defaults to None.

    Returns:
        pd.DataFrame: pandas dataframe of the ECDF data.
    """
    if x_values is None:
        if x_min is None:
            x_min = data[eval_var].min()
        if x_max is None:
            x_max = data[eval_var].max()
        x_values = get_sequence(
            x_min, x_max, 50, scale_log=scale_xlog, cast_to_int=True
        )
    data_aligned = align_data(
        data.cast({eval_var: pl.Int64}),
        x_values,
        group_cols=["data_id"],
        x_col=eval_var,
        y_col=fval_var,
        maximization=maximization,
    )
    dt_ecdf = (
        transform_fval(
            data_aligned,
            fval_col=fval_var,
            maximization=maximization,
            lb=y_min,
            ub=y_max,
            scale_log=scale_ylog,
        )
        .group_by([eval_var] + free_vars)
        .mean()
        .sort(eval_var)
    ).to_pandas()
    return dt_ecdf

def get_trajectory(data: pl.DataFrame, 
                   traj_length: int = None,
                   min_fevals: int = 1,
                   evaluation_variable: str = "evaluations",
                   fval_variable: str = "raw_y",
                   free_variables: Iterable[str] = ["algorithm_name"],
                    maximization: bool = False
) -> pl.DataFrame:
    """get the trajectory of the performance of the algorithms in the data
    This function aligns the data to a fixed number of evaluations and returns the performance trajectory.

    Args:
        data (pl.DataFrame): The DataFrame resulting from loading the data from a DataManager.
        traj_length (int, optional): Length of the trajecotry. Defaults to None.
        min_fevals (int, optional): Evaluation number from which to start the trajectory. Defaults to 1.
        evaluation_variable (str, optional): Variable corresponding to evaluation count in `data`. Defaults to "evaluations".
        fval_variable (str, optional): Variable corresponding to function value in `data`. Defaults to "raw_y".
        free_variables (Iterable[str], optional): Free variables in `data`. Defaults to ["algorithm_name"].
        maximization (bool, optional): Whether the data is maximizing or not. Defaults to False.

    Returns:
        pd.DataFrame: DataFrame: A polars DataFrame with the aligned data, where each row corresponds to a specific evaluation count and the performance value.
    """
    if traj_length is None:
        max_fevals = data[eval_var].max()
    else:
        max_fevals = traj_length + min_fevals
    x_values = np.arange(min_fevals, max_fevals + 1) 
    data_aligned = align_data(
        data.cast({evaluation_variable: pl.Int64}),
        x_values,
        group_cols=["data_id"] + free_variables,
        x_col=evaluation_variable,
        y_col=fval_variable,
        maximization=maximization,
    )
    return data_aligned